from selenium import webdriver

class Remote(object):
    name = None
    url_template = None

    def __init__(self, conf):
        self.conf = conf
        self.driver = None
        self.tunnel = None
        self.tunnel_id = None

    def build_driver(self, capabilities):
        self.driver = webdriver.Remote(
            desired_capabilities=capabilities,
            command_executor=self.url_template.format(
                credentials=self.credentials))
        return self.driver

    def sanitize_config(self, config):
        return config

    @property
    def credentials(self):
        raise NotImplementedError()

    def get_unused_port(self):
        raise NotImplementedError()

    def set_test_status(self, passed=True):
        raise NotImplementedError()

    def start_tunnel(self):
        raise NotImplementedError()

    def set_tunnel_id(self, tunnel_id):
        self.tunnel_id = tunnel_id

    def stop_tunnel(self):
        raise NotImplementedError()



from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select
from selenium.common.exceptions import StaleElementReferenceException

class Table(object):
    name = None
    cssid = None
    initialized_locator = None
    field_selectors = None

    def __init__(self, util, name=None, cssid=None,
                 initialized_locator=None, field_selectors=None):
        """
        :param util: The util object to use to access Selenium.

        :type util: :class:`selenic.util.Util`

        :param name: The user-friendly name to give to the table.

        :param cssid: The CSS id to use to find the table.

        :param initialized_locator: A selector to use to detect
        whether the table is ready.

        :type initialized_locator: This should be a ``(By..., string)`` pair.

        :param field_selectors: A list of CSS selectors to scope the
        field searches. When the class searches for fields, it will
        search in each of the CSS selectors passed here. We do not
        verify that these selectors are within the table selected by
        ``cssid``. The fields may thus be *outside* the table.

        :type field_selectors: A list of strings.
        """
        self.util = util

        if cssid:
            self.cssid = cssid

        if name:
            self.name = name

        if initialized_locator:
            self.initialized_locator = None

        if field_selectors:
            self.field_selectors = field_selectors

        if self.field_selectors is None:
            self.field_selectors = []

        self._found_fields = None

    def add_field_selector(self, selector):
        self.field_selectors.append(selector)

    @property
    def _search_fields(self):
        if not self._found_fields:
            self._find_search_fields()

        return self._found_fields

    def _find_search_fields(self):
        self.wait_for_initialized()

        found = self.util.driver.execute_script("""
        var cssid = arguments[0];
        var selectors = arguments[1];
        var found = {};

        function search_selector(selector) {
          var scope = document.querySelector(selector);
          var types = ["input", "select"];
          for (var type_ix = 0, type; (type = types[type_ix]); ++type_ix) {
            var els = scope.getElementsByTagName(type);
            for (var el_ix = 0, el; (el = els[el_ix]); ++el_ix) {
              var label = el.closest("label");
              if (!label)
                throw new Error("no label for " + el);

              // We clone the label, and clean out the form controls.
              // ``select`` elements in particular will produce text
              // due to the  ``option`` elements.

              var clone = label.cloneNode(true);
              var child = clone.firstElementChild;
              while (child) {
                var next = child.nextElementSibling;
                if (types.indexOf(child.localName) > -1) {
                  child.parentNode.removeChild(child);
                }
                child = next;
              }

              label = clone.textContent.trim().replace(/:$/, '');
              found[label] = el;
            }
          }
        }

        for (var selector_ix = 0, selector; (selector = selectors[selector_ix]);
          ++selector_ix) {
            search_selector(selector);
        }

        // We cannot just return the dictionary we populated, because Selenium
        // won't convert the values to proper WebElements. Flattening
        // the dictionary like this does the trick.

        var ret = [];
        var keys = Object.keys(found);
        for (var key_ix = 0, key; (key = keys[key_ix]); ++key_ix) {
            ret.push(key, found[key]);
        }
        return ret;
        """, self.cssid, self.field_selectors)
        # Reconstitute the dictionary
        self._found_fields = dict(zip(found[::2], found[1::2]))

    def call_with_search_field(self, name, callback):
        """
        Calls a piece of code with the DOM element that corresponds to
        a search field of the table.

        If the callback causes a ``StaleElementReferenceException``,
        this method will refetch the search fields and try
        again. Consequently **the callback should be designed to be
        callable multiple times and should only interact with the
        search field passed to it.** It should not fetch or interact
        with other DOM elements.

        :param name: The name of the field to use.
        :param callback: The callback to call. The first parameter
        will be the Selenium ``WebElement`` that is the search field.
        """
        done = False
        while not done:
            def check(*_):
                if name in self._search_fields:
                    return True

                # Force a refetch
                self._found_fields = None
                return False

            self.util.wait(check)
            field = self._search_fields[name]
            try:
                callback(field)
                done = True
            except StaleElementReferenceException:
                # We force a refetch of the fields
                self._found_fields = None

    def wait_for_initialized(self):
        """
        This code will wait using the currently active timeout set on
        the ``Util`` object that was passsed when this instance was
        created.
        """

        # Waiting for the existence of a .dataTable element is not
        # reliable. We wait for the existence of the "next" button
        # instead.
        return len(self.util.find_elements(self.initialized_locator)) > 0

    def fill_field(self, name, value):
        driver = self.util.driver

        def fill(el):
            driver.execute_script("arguments[0].value = arguments[1];",
                                  el, value[:-1])
            self.setup_redraw_check()
            el.send_keys(value[-1])
            self.wait_for_redraw()

        self.call_with_search_field(name, fill)

    def set_select_option(self, name, value):
        def select(el):
            select = Select(el)
            self.setup_redraw_check()
            select.select_by_visible_text(value)
            self.wait_for_redraw()

        self.call_with_search_field(name, select)

    def setup_redraw_check(self):
        raise NotImplementedError()

    def wait_for_redraw(self):
        raise NotImplementedError()

    def wait_for_results(self, expected_total):
        raise NotImplementedError()

    def get_result(self, number):
        raise NotImplementedError()



import re
import os
import subprocess
import types
from distutils.version import StrictVersion

import selenium
from selenium import webdriver
from selenium.webdriver.firefox.webdriver import FirefoxProfile, FirefoxBinary
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

from . import remote, outil
from .capabilities import NormalizedCapabilities

CHROMEDRIVER_ELEMENT_CENTER_PATCH_FLAG = \
    "_selenic_chromedriver_element_center_patch"

class Builder(object):

    def __init__(self, config_path, options):
        """
        Initializes a configuration.

        :param config_path: The configuration file to use. Must be a valid
                            Python file.
        :type config_path: :class:`basestring`
        :param options: A dictionary of key/value pairs with which the
                        global variable ``builder_args`` will be initialized
                        before the configuration is read.
        """
        self.config_path = config_path

        self.local_conf = {
            'builder_args': options
        }
        execfile(self.config_path, self.local_conf)

        # This effectively lets the user force colon handling on or
        # off if COLON_HANDLING is defined.
        colon_handling = self.local_conf.get("COLON_HANDLING", None)

        self.config = self.local_conf.get("CONFIG")

        if colon_handling is None:
            self.colon_handling = self.config.browser == \
                "INTERNETEXPLORER" and \
                StrictVersion(
                    selenium.__version__) <= StrictVersion("2.40.0")
        elif colon_handling is True:
            self.colon_handling = True
        elif colon_handling is False:
            self.colon_handling = False
        else:
            raise Exception("bad value for COLON_HANDLING: " + colon_handling)

        self.remote = self.config.remote
        self.remote_service = None

        if self.remote:
            remote_service = self.local_conf["REMOTE_SERVICE"]
            self.remote_service = \
                remote.get_service_cls(remote_service)(self.local_conf)

    def __getattr__(self, name):
        if name in self.local_conf:
            return self.local_conf[name]

        raise AttributeError("{!r} object has no attribute {!r}"
                             .format(self.__class__, name))

    def get_driver(self, desired_capabilities=None):
        """
        Creates a Selenium driver on the basis of the configuration file
        upon which this object was created.

        :param desired_capabilities: Capabilities that the caller
            desires to override. This have priority over those
            capabilities that are set by the configuration file passed
            to the builder.
        :type desired_capabilities: class:`dict`
        :returns: A driver.
        :raises ValueError: When it can't figure out how to create a
                            browser as specified by the BROWSER
                            configuration variable.
        """
        override_caps = desired_capabilities or {}

        desired_capabilities = \
            self.config.make_selenium_desired_capabilities()
        desired_capabilities.update(override_caps)
        browser_string = self.config.browser

        chromedriver_version = None
        if self.remote:
            driver = self.remote_service.build_driver(desired_capabilities)
            # There is no equivalent for BrowserStack.
            if browser_string == "CHROME" and \
               self.remote_service.name == "saucelabs":
                chromedriver_version = \
                    desired_capabilities.get("chromedriver-version", None)
                if chromedriver_version is None:
                    raise ValueError(
                        "when using Chrome, you must set a "
                        "``chromedriver-version`` capability so that Selenic "
                        "can detect which version of Chromedriver will "
                        "be used.")
        else:
            if browser_string == "CHROME":
                chromedriver_path = self.local_conf["CHROMEDRIVER_PATH"]
                driver = webdriver.Chrome(
                    chromedriver_path,
                    chrome_options=self.local_conf.get("CHROME_OPTIONS"),
                    desired_capabilities=desired_capabilities,
                    service_log_path=self.local_conf["SERVICE_LOG_PATH"],
                    service_args=self.local_conf.get("SERVICE_ARGS"))
                version_line = subprocess.check_output(
                    [chromedriver_path, "--version"])
                version_str = re.match(ur"^ChromeDriver (\d+\.\d+)",
                                       version_line).group(1)
                chromedriver_version = StrictVersion(version_str)
            elif browser_string == "FIREFOX":
                profile = self.local_conf.get("FIREFOX_PROFILE") or \
                    FirefoxProfile()
                binary = self.local_conf.get("FIREFOX_BINARY") or \
                    FirefoxBinary()
                driver = webdriver.Firefox(profile, binary,
                                           capabilities=desired_capabilities)
            elif browser_string == "INTERNETEXPLORER":
                driver = webdriver.Ie()
            elif browser_string == "OPERA":
                driver = webdriver.Opera()
            else:
                # SAFARI
                # HTMLUNIT
                # HTMLUNITWITHJS
                # IPHONE
                # IPAD
                # ANDROID
                # PHANTOMJS
                raise ValueError("can't start a local " + browser_string)

            # Check that what we get is what the config wanted...
            driver_caps = NormalizedCapabilities(driver.desired_capabilities)
            browser_version = \
                re.sub(r"\..*$", "", driver_caps["browserVersion"])

            if driver_caps["platformName"].upper() != self.config.platform:
                raise ValueError("the platform you want is not the one "
                                 "you are running selenic on")

            if browser_version != self.config.version:
                raise ValueError("the version installed is not the one "
                                 "you wanted")

        # On BrowserStack we cannot set the version of chromedriver or
        # query it. So we make the reasonable assuption that the
        # version of chromedriver is greater than 2.13. (There have
        # been at least 7 releases after 2.13 at the time of writing.)
        if (self.remote_service and
            self.remote_service.name == "browserstack") or \
           (chromedriver_version is not None and
                chromedriver_version > StrictVersion("2.13")):
            # We patch ActionChains.
            chromedriver_element_center_patch()
            # We need to mark the driver as needing the patch.
            setattr(driver, CHROMEDRIVER_ELEMENT_CENTER_PATCH_FLAG, True)

        driver = self.patch(driver)
        return driver

    def update_ff_binary_env(self, variable):
        """
        If a ``FIREFOX_BINARY`` was specified, this method updates an
        environment variable used by the ``FirefoxBinary`` instance to
        the current value of the variable in the environment.

        This method is a no-op if ``FIREFOX_BINARY`` has not been
        specified or if the configured browser is not Firefox.

        A common use-case for this method is updating ``DISPLAY`` once
        an Xvfb or Xephyr instance has been launched. Typically, by
        the time these displays are launched, the configuration file
        has already been loaded and whatever ``FirefoxBinary``
        instance was created for ``FIREFOX_BINARY`` has a stale
        ``DISPLAY`` value.

        :param variable: The name of the variable to update.
        :type variable: :class:`str`
        """
        if self.config.browser != 'FIREFOX':
            return

        binary = self.local_conf.get('FIREFOX_BINARY')
        if binary is None:
            return

        # pylint: disable=protected-access
        binary._firefox_env[variable] = os.environ[variable]

    def set_test_status(self, passed=True):
        if not self.remote_service:
            return

        self.remote_service.set_test_status(passed)

    def get_unused_port(self):
        return outil.get_unused_port() if not self.remote_service else \
            self.remote_service.get_unused_port()

    def start_tunnel(self):
        if not self.remote_service:
            return None

        return self.remote_service.start_tunnel()

    def set_tunnel_id(self, tunnel_id):
        if not self.remote_service:
            return None

        self.remote_service.set_tunnel_id(tunnel_id)

    def stop_tunnel(self):
        if not self.remote_service:
            return None

        return self.remote_service.stop_tunnel()

    def patch(self, driver):
        if self.colon_handling:
            import types
            driver.find_element = types.MethodType(make_patched_find_element(
                driver.find_element), driver)
            driver.find_elements = types.MethodType(make_patched_find_element(
                driver.find_elements), driver)

            WebElement.find_element = \
                make_patched_find_element(WebElement.find_element)
            WebElement.find_elements = \
                make_patched_find_element(WebElement.find_elements)

        return driver


def make_patched_find_element(original):

    def method(self, by=By.ID, value=None):
        if By.is_valid(by) and by == By.CLASS_NAME:
            by = By.CSS_SELECTOR
            value = "." + value

        if original.im_self:
            return original(by, value)
        else:
            return original(self, by, value)
    return method


def chromedriver_element_center_patch():
    """
    Patch move_to_element on ActionChains to work around a bug present
    in Chromedriver 2.14 to 2.20.

    Calling this function multiple times in the same process will
    install the patch once, and just once.
    """

    patch_name = "_selenic_chromedriver_element_center_patched"
    if getattr(ActionChains, patch_name, None):
        return  # We've patched ActionChains already!!

    # This is the patched method, which uses getBoundingClientRect
    # to get the location of the center.
    def move_to_element(self, el):
        pos = self._driver.execute_script("""
        var rect = arguments[0].getBoundingClientRect();
        return { x: rect.width / 2, y: rect.height / 2};
        """, el)
        self.move_to_element_with_offset(el, pos["x"], pos["y"])
        return self

    old_init = ActionChains.__init__

    def init(self, driver):
        old_init(self, driver)

        # Patch the instance, only if the driver needs it.
        if getattr(driver, CHROMEDRIVER_ELEMENT_CENTER_PATCH_FLAG, None):
            self.move_to_element = types.MethodType(move_to_element, self)

    ActionChains.__init__ = init

    # Mark ActionChains as patched!
    setattr(ActionChains, patch_name, True)



import os
import tempfile
import subprocess
import signal
import base64
import httplib
import time
try:
    import json
except ImportError:
    import simplejson as json

from .base import Remote
from ..outil import get_unused_port

def set_test_status(jobid, credentials, passed=True):
    """
    Sets the final status for test runs that are executed at
    SauceLabs.

    :param jobid: The job id for which to set the status.
    :type jobid: :class:`basestring`
    :param credentials: The "user:key" string.
    :type credentials: :class:`basestring`
    :param passed: Whether or not the test run passed. Defaults to ``True``.
    :type passed: :class:`bool`
    :raises Exception: When it can't set the status.
    """

    # curl -u "user:key"
    #      -X PUT
    #      -H "Content-Type: application/json"
    #      -d "{\"status\":\"<new-status>\", \"reason\":\"<reason text>\"}"
    #      https://www.browserstack.com/automate/sessions/<session-id>.json

    (username, key) = credentials.split(":")
    creds = base64.encodestring('%s:%s' % (username, key))[:-1]

    conn = httplib.HTTPSConnection("www.browserstack.com")
    conn.request('PUT', '/automate/sessions/%s.json' % jobid,
                 json.dumps({"status": "completed" if passed else "error"}),
                 headers={"Content-Type": "application/json",
                          "Authorization": "Basic " + creds})
    resp = conn.getresponse()
    if resp.status != 200:
        raise Exception("got response: " + resp.status)

class Tunnel(object):

    def __init__(self, path, key):
        self.path = path
        self.key = key
        self.process = None
        self.tunnel_id = "bslocal-for-" + str(os.getpid())
        self.tmpdir = None

    def start(self):
        self.tmpdir = tmpdir = tempfile.mkdtemp()
        stdout_path = os.path.join(tmpdir, "stdout")
        tunnel_id = self.tunnel_id
        #
        # We use os.setsid to create a new process group. We want this
        # because BrowserStackLocal creates children that won't die if
        # we just kill the parent.
        #
        self.process = subprocess.Popen(
            [self.path, self.key, "-onlyAutomate",
             "-localIdentifier", tunnel_id],
            stdin=open("/dev/null", 'r'),
            stdout=open(stdout_path, 'w'),
            preexec_fn=os.setsid)

        # This is awful but less awful that other
        # alternatives. BrowserStackLocal does not have a well-defined
        # mechanism to know when it is ready so we have to
        # periodically check its stdout. We do not use a pipe due to
        # the well-known issue with reading pipes without allocating a
        # thread to continually read it and prevent the buffer being
        # filled.
        with open(stdout_path, 'r') as stdout:
            while True:
                if "Press Ctrl-C to exit" in stdout.read():
                    break
                time.sleep(0.2)

        return tunnel_id

    def stop(self):
        if self.process:
            # SIGTERM does not do it...
            os.killpg(os.getpgid(self.process.pid), signal.SIGKILL)
            self.process = None


class BrowserStack(Remote):
    name = "browserstack"
    url_template = "http://{credentials}@hub.browserstack.com:80/wd/hub"

    def __init__(self, *args, **kwargs):
        super(BrowserStack, self).__init__(*args, **kwargs)
        self.tunnel = None

    @property
    def credentials(self):
        return self.conf["BROWSERSTACK_CREDENTIALS"]

    def build_driver(self, capabilities):
        caps = sanitize_capabilities(capabilities)

        if self.tunnel_id and self.tunnel:
            raise Exception("you have set a tunnel id and have "
                            "requested a tunnel; you should do one or the "
                            "other")

        if self.tunnel_id or self.tunnel:
            caps['browserstack.local'] = True
            caps['browserstack.localIdentifier'] = \
                self.tunnel_id or self.tunnel.tunnel_id

        return super(BrowserStack, self).build_driver(capabilities)

    def get_unused_port(self):
        return get_unused_port()

    def start_tunnel(self):
        _, key = self.credentials.split(":")
        self.tunnel = Tunnel(self.conf["BSLOCAL_PATH"], key)
        return self.tunnel.start()

    def stop_tunnel(self):
        self.tunnel.stop()

    def set_test_status(self, passed=True):
        """
        Sets the final status for test runs that are executed at
        SauceLabs.

        :param passed: Whether or not the test run passed.
        :type passed: :class:`bool`
        :raises Exception: When it can't set the status.
        """
        set_test_status(self.driver.session_id, self.credentials, passed)


def sanitize_capabilities(caps):
    """
    Sanitize the capabilities we pass to Selenic so that they can
    be consumed by Browserstack.

    :param caps: The capabilities passed to Selenic. This dictionary
    is modified.

    :returns: The sanitized capabilities.
    """
    platform = caps["platform"]

    upper_platform = platform.upper()

    if upper_platform.startswith("WINDOWS 8"):
        caps["platform"] = "WIN8"
    elif upper_platform.startswith("OS X "):
        caps["platform"] = "MAC"
    elif upper_platform == "WINDOWS 10":
        del caps["platform"]
        caps["os"] = "Windows"
        caps["os_version"] = "10"

    if caps["browserName"].upper() == "MICROSOFTEDGE":
        # Sauce Labs takes complete version numbers like
        # 15.1234. However, Browser Stack takes only .0 numbers like
        # 15.0.
        caps["version"] = caps["version"].split(".", 1)[0] + ".0"

    caps["browser_version"] = caps["version"]
    del caps["version"]

    return caps
